/*************************************************************************************************************************************************
 * Nano slave servo driver
 *
 *************************************************************************************************************************************************/

#include "Servo.h"

#define SERVO_FREQ 50 // 50 Hz updates

const byte NumServos = 6;
const int VelPeriod = 50; // how often the servo position is recalculated
const int USMIN[NumServos] = {600,600,600,600,600,600}; // minimum pulse length (in uS)
const int USMAX[NumServos] = {2400,2400,2400,2400,2400,2400}; // maximum pulse length (in uS)
      
Servo myservo[NumServos];

const int ServoPins[NumServos] = { 3, 5, 6, 9, 10, 11 };

const int InitialPosition[NumServos] = { // pulse len in microseconds
 (USMAX[0] + USMIN[0]) / 2,
 (USMAX[1] + USMIN[1]) / 2,
 (USMAX[2] + USMIN[2]) / 2,
 (USMAX[3] + USMIN[3]) / 2,
 (USMAX[4] + USMIN[4]) / 2,
 (USMAX[5] + USMIN[5]) / 2};

float target[NumServos]; // target position
float p[NumServos]; // current position
float v[NumServos]; // velocity: (change in p per step)
float vmax[NumServos]; // max change in p per step
float amax[NumServos];   // max change in v per step
byte group[NumServos];   // 0 = move independently; 1..n = move together

bool AtTargets = true;

//-----------------------------------------------------------
// recvByte()
//  wait for a byte from the PC
//-----------------------------------------------------------
byte recvByte() {
  while (! Serial.available()) ;
  byte b = Serial.read();
  return b;
}

//-----------------------------------------------------------
// recvCmd()
//   get command byte from the PC and act on it
//-----------------------------------------------------------
void recvCmd() {
  if (Serial.available() > 0) {
    int b = Serial.read();
    execCmd(b);
  }
}

//-----------------------------------------------------------
// execCmd(int b)
//   execute command byte 
//     a complete command is low bits followed by high bits    
//       0nnnnnnn   n = low 7 bits
//       10mmmsss   m = high 3 bits: set target of servo sss to mmmnnnnnnn0 + 500 microseconds (500..2547)
//       1100msss   m = high bit: set max velocity of servo sss to mnnnnnnn/scale microseconds per interval
//       1101msss   m = high bit: set max acceleration of servo sss to mnnnnnnn/scale microseconds per interval
//       11100sss   servo sss is in group n
//
//     if acceleration = 0 then goes instantly to max velocity 
//     if velocity = 0 then goes instantly to target
//     if two or more targets are set then adjusts acc/vel so both arrive at same time
//
//    examples
//        7A 8D fully CW
//        6E AD centre
//        36 BD fully CCW
//        0F C5 slow
//        7F CD fast
//-----------------------------------------------------------
void execCmd(int b) {
  static int lowBits = 0;

  if (b < 0b10000000)
    lowBits = b; // low 7 bits
  else {
    byte servo = min(b & 0b111, NumServos-1);
    int m = (b >> 3) & 0b111;
    lowBits = (m << 7) + lowBits;
    
    if ((b & 0b11000000) == 0b10000000) { // set target
      target[servo] = lowBits*2 + 500;
      target[servo] = constrain(target[servo], USMIN[servo], USMAX[servo]);
    } else {
      lowBits &= 255;
      switch (b & 0b11110000) {
        case 0b11000000: vmax[servo] = lowBits; if (vmax[servo] == 0) vmax[servo] = USMAX[servo]-USMIN[servo]; break; // max velocity = mnnnnnnn
        case 0b11010000: amax[servo] = lowBits/10.0; break; // max acceleration = mnnnnnnn/10
        case 0b11100000: group[servo] = lowBits; break; // group = nnnnnnn
        case 0b11110000:  // for debugging
          break;
      }
    }
  }
}

//-----------------------------------------------------------
// sign()
//   sign of number
//   a function not a macro so you don't calulate the parameter twice
//-----------------------------------------------------------
float sign(float a) {
  if (a < 0) return -1;
  if (a > 0) return +1;
  return 0;
}

//-----------------------------------------------------------
// sqr()
//   square funtion
//   it's a function not a macro like sq() 
//-----------------------------------------------------------
float sqr(float a) {
  return a*a;
}

//-----------------------------------------------------------
// MoveServos()
//   change servo position so they're closer to their target
//-----------------------------------------------------------
void MoveServos() {
  static unsigned long previousMillis = 0;
  if (millis() - previousMillis < VelPeriod)
    return;
  previousMillis = millis();

  float g[NumServos];
  float v1,a1,tmax,sign1,p0;

  for (byte grp=0; grp < NumServos; grp++) {
    tmax = 0;
    for (byte servo=0; servo < NumServos; servo++) 
    if (group[servo] == grp) {
      g[servo] = abs(target[servo]-p[servo])/vmax[servo];
      if (amax[servo] > 0) {
        sign1 = sign(sqr(v[servo]) + amax[servo]*2*(2*abs(v[servo]) + sign(v[servo])*(p[servo]-target[servo])));
        g[servo] = max(g[servo],
          (sign1*abs(v[servo]) + sqrt(abs(2*sign1*v[servo]*abs(v[servo]) + 4*amax[servo]*(p[servo]-target[servo]))))/amax[servo]);
      }
      tmax = max(tmax,g[servo]);
    }
  
    for (byte servo=0; servo < NumServos; servo++) 
    if (group[servo] == grp) {    
      if (grp == 0) 
        tmax = g[servo];
      if ((tmax < 2) or ((abs(v[servo]) <= amax[servo]+0.0001) and (abs(p[servo]+v[servo]-target[servo]) <= 1))) {
        p[servo] = target[servo];
        v[servo] = 0;
      } else
      if (tmax > 0) {
        g[servo] = g[servo]/tmax;
        a1 = amax[servo]*sqr(g[servo]);
  
        if (a1 == 0) {
          v[servo] = (target[servo]-p[servo])/tmax;
        } else {
          v1 = vmax[servo]*g[servo];
          p0 = 2*v[servo] + abs(v[servo])*v[servo]/(2*a1);
          a1 = sign(target[servo]-p[servo]-p0)*a1;
          if (a1*(target[servo]-p[servo]) < 0) {
            a1  =  sign(a1)*abs(v[servo])*v[servo]/(2*(target[servo]-p[servo]));
            a1 = constrain(a1,-amax[servo],amax[servo]);
          }
  
          v[servo] = v[servo]+a1;
          v[servo] = constrain(v[servo],-v1,v1);
        }
        
        v[servo] = constrain(v[servo],-vmax[servo],vmax[servo]);
        p[servo] = p[servo]+v[servo];
      }
    }
  }

  for (byte servo=0; servo < NumServos; servo++) {    
    p[servo] = constrain(p[servo],USMIN[servo],USMAX[servo]);
    myservo[servo].writeMicroseconds(p[servo]);
  }
}

//-----------------------------------------------------------
// TestSeq()
//   a test sequence for demo
//-----------------------------------------------------------
void SetTarg(byte servo, int pos, int tWait) {
  if (Serial.available()) 
    return;
    
  target[servo] = pos; 

  if (tWait > 0) {
    long u = millis();
    while (millis() < u+tWait) {
      MoveServos();    
    }
  }
}

void TestSeq() {
  vmax[0] = 2000; amax[0] = 0;    group[0] = 1;
  vmax[1] = 2000; amax[1] = 0;    group[1] = 1;
  vmax[2] = 2000; amax[2] = 10.0; group[2] = 0;
  vmax[3] = 2000; amax[3] = 0;    group[3] = 0;
  vmax[4] = 2000; amax[4] = 10.0; group[4] = 1;
  vmax[5] = 2000; amax[5] = 10.0; group[5] = 1;

  while (! Serial.available()) {
    SetTarg(0,1500,0); SetTarg(1,1500,0); SetTarg(2,1750,0); SetTarg(3,1500,0); SetTarg(4,1600,0); SetTarg(5,1500,500);
    SetTarg(0,1000,500);
    SetTarg(0,1900,500);
    SetTarg(0,1496,500);
    SetTarg(1,2200,500);
    SetTarg(1,1200,500);
    SetTarg(0,1200,0); SetTarg(1,1200,500);
    SetTarg(0,1700,500);
    SetTarg(1,2200,500);
    SetTarg(0,1200,500);
    SetTarg(0,1500,0); SetTarg(1,1500,1000);
    SetTarg(4,1400,0); SetTarg(5,500,0); SetTarg(3,1300,1000);
    SetTarg(4,2000,0); SetTarg(3,1700,1000);
    SetTarg(5,2300,1000);
    SetTarg(4,1300,0); SetTarg(3,1300,1000);
    SetTarg(4,1600,0); SetTarg(5,1500,0); SetTarg(3,1500,1000);
    SetTarg(0,1886,0); SetTarg(4,1600,1000);
    SetTarg(0,1476,0); SetTarg(5,2012,1000);
    SetTarg(0,1500,0); SetTarg(5,1500,1000);
    SetTarg(0,1000,0); SetTarg(5,2012,1000);
    SetTarg(0,1004,0); SetTarg(1,1500,0); SetTarg(2,1500,0); SetTarg(3,1500,0); SetTarg(4,1600,0); SetTarg(5,1500,2000);
    SetTarg(0,1461,0); SetTarg(5,752,2000);
    SetTarg(0,1500,0); SetTarg(5,1500,2000);
    SetTarg(0,1156,0); SetTarg(1,1831,0); SetTarg(2,1000,0); SetTarg(3,1621,0); SetTarg(4,1796,0); SetTarg(5,752,2000);
    SetTarg(5  ,800,0); SetTarg(03,1000,0); SetTarg(0,1500,0); SetTarg(1,1500,200);
    SetTarg(5,1100,0); SetTarg(03,1900,200);
    SetTarg(5,1400,0); SetTarg(03,1000,200);
    SetTarg(5,1700,0); SetTarg(03,1900,200);
    SetTarg(5,2000,0); SetTarg(03,1000,200);
    SetTarg(5,2300,0); SetTarg(03,1900,200);
    SetTarg(5,2500,0); SetTarg(03,1000,200);
    SetTarg(3,1500,0); SetTarg(5,1500,2000);
    SetTarg(0,1500,0); SetTarg(1,1500,0); SetTarg(2,1750,0); SetTarg(3,1500,0); SetTarg(4,1600,0); SetTarg(5,1500,5000);

    SetTarg(3,1000,200);
    SetTarg(0,1500,0); SetTarg(1,1500,0); SetTarg(2,1750,0); SetTarg(3,1500,0); SetTarg(4,1600,0); SetTarg(5,1500,2000);
  }
}

//-----------------------------------------------------------
// setup() 
//-----------------------------------------------------------
void setup() 
{
  Serial.begin(19200);
  Serial.println("Booting..."); 

  delay(10);

  for (byte servo = 0; servo < NumServos; servo++) {
    p[servo] = InitialPosition[servo];
    v[servo] = 0;
    target[servo] = p[servo];
    amax[servo] = 0.1; 
    vmax[servo] = 10.0; 
    myservo[servo].writeMicroseconds(p[servo]);
    myservo[servo].attach(ServoPins[servo]);
    group[servo] = 0;
  }

  Serial.println("Initialised"); 

//  TestSeq();
}

void loop() {
  recvCmd();
  MoveServos();
}
